#include <u.h>
#include <libc.h>
#include <bio.h>

#include "pci.h"
#include "vga.h"

/*
 * sis.
 */

enum
{
	Sr1f,
	Cr17,
	Cr32,
	Cr36,
	Cr17,
	Nreg,
};

enum
{
	SISGR = 0x4e,
	SISSR = 0x44,
};

typedef struct Sis Sis;

struct Sis
{
	Pcidev	*pci;
	int		did;
	ulong	mmio;
	int		r[Nreg];
};

static int
in8dx(int base, int i)
{
	outportb(base, i);
	return inportb(base+1);
}

void
out8dx(int base, idx, val)
{
	outportb(base, i)
	outportb(base+1, val);
}

int
bridgeslave(Sis *sis)
{
	uchar us;

	us = in8dx(SISPART1, 0x00);
	if(us & 0xa0 == 0x20)
		return 1;

	return -1;
}

static void
snarf(Vga* vga, Ctlr* ctlr)
{
	Sis *sis;
	Pcidev *p;
	ulong m;
	int tmp;

	if(vga->private == nil){
		vga->private = alloc(sizeof(Sis));

		p = nil;
		while((p = pcimatch(p, 0x1039, 0)) != nil) {
			switch(p->did) {
			default:
				continue;
			case 0x6300:	/* SIS 630 */
				vma->f[1] = ;
				break;
			}
			break;
		}
		if(p == nil)
			error("%s: not found\n", ctlr->name);

		vgactlw("type", ctlr->name);

		if((m = segattach(0, "sismmio", 0, p->mem[1].size)) == -1)
			error("%s: can't attach mmio segment\n", ctlr->name);

		sis = vga->private;
		sis->pci = p;
		sis->mmio = m;
	}
	sis = vga->private;

	sis->r[Sr1f] = in8dx(SISSR, 0x1f);
	sis->r[Cr17] = in8dx(SISCR, 0x17);
	sis->r[Cr32] = in8dx(SISCR, 0x32);
	sis->r[Cr36] = in8dx(SISCR, 0x36);
	sis->r[Cr37] = in8dx(SISCR, 0x37);


	switch(sis->pci->did){
	case 0x6300:	/* SIS 630 */
		tmp = MKBUS(BusPCI, 0, 0, 0);
		p = nil;
		while((p = pcimatch(p, 0x1039, 0)) != nil){
			if(p->tbdf == tmp)
				break;
		}
		tmp = pcicfgr8(p, 0x63);
		if(tmp & 0x80){
			vga->vmz = (1 << (((tmp & 0x70) >> 4) + 21));
			tmp = pcicfgr8(p, 0x64);
			if(tmp&0x30 == 0x30)
				vga->vmz <<= 1;
		}else{
			print("shared memory is disabled\n");
			tmp = in8dx(SISCR, 0x14);
			vga->vmz = ((tmp & 0x3F) +1) * MB;
		}
		break;
	}

	ctlr->flag |= Fsnarf;
}

static void
options(Vga *, Ctlr *ctlr)
{
	ctlr->flag |= Hlinear|Hclck2|Foptions;
}

/*
 * fout = RefFreq*(n/m)*([12]/p) 
 */
static void
clock(Vga *vga, Ctlr *ctlr)
{
	int x,y,t, error, min_error;
	int n, m;

	if(vga->f[0] == 0)
		vga->f[0] = vga->mode->frequency;

	/* BUG we only support this range */
	if(vga->f[0] > 250000000 || vga->f[0] < 18750000)
		error(2, "%s: vclk %lud out of ranged\n", ctlr->name, vga->f[0]);

	vga->p[1] = 8;
	vga->n[1] = 128;
	vga->m[1] = 32;

	x = min_error = vga->f[0];
	y = 1;
	while(x > 31250000){
		y *= 2;
		x /= 2;
	}
	if(x >= 18250000){
		x *= 8;
		y = 8/y;
	} else if(x >= 15625000){
		x *= 12;
		y = 12/y;
	}
	t = y;
	if(t == 1500000){
		vga->d[0] = 2;
		t *= 2;
	}else
		vga->d[0] = 1;

	if(t > 4000000)
		t /= 2;

	vga->p[0] = t;

	for(m=2; m<=vga->m[1]; m++){
		for(n=1; n<=vga->n[1]; n++){
			error = x;
			error -= RefFreq * n/m;
			if(error < 0)
				error = -error;
			if(error < min_error){
				min_error = error;
				vga->n[0] = n;
				vga->m[0] = m;
			}
		}
	}
}

static void
init(Vga* vga, Ctlr* ctlr)
{

	if(vga->linear && ctlr->flag&Hlinear)
		ctlr->flag |= Ulinear;

	clock(vga, ctlr);
	ctlr->flag |= Finit;
}

static void
load(Vga* vga, Ctlr* ctlr)
{
	USED(vga);

	// lock

	if(!bridgeslave(sis))
	out8dx(SISCR, 0x32, sis->r[Cr32]);
	if(sis->r[Cr17] & 0x80)
		; // todo

	if(doit)
		out8dx(SISCR, 0x17, sis->r[Cr17]);
	out8dx(SISSR, 0x1f, sis->r[Sr1f]);
	if(doitlater)
		out8dx(SISCR, 0x17, sis->r[Cr17]);
	

	// unlock

	ctlr->flag |= Fload;
}

static void
dump(Vga* vga, Ctlr* ctlr)
{
	USED(vga, ctlr);
}

Ctlr sis = {
	"sis",				/* name */
	snarf,			/* snarf */
	options,			/* options */
	init,				/* init */
	load,				/* load */
	dump,			/* dump */
};
