#include "u.h"
#include "../port/lib.h"
#include "mem.h"
#include "dat.h"
#include "fns.h"
#include "io.h"
#include "../port/error.h"

#define	Image	IMAGE
#include <draw.h>
#include <memdraw.h>
#include <cursor.h>
#include "screen.h"


static Pcidev*
sispci(void)
{
	Pcidev *p;

	p = nil;
	while((p = pcimatch(p, 0x1039, 0)) != nil){
		switch(p->did){
		default:
			continue;
		case 0x6300:		/* sis 630 */
			return p;
		}
	}
	return nil;
}


static ulong
sislinear(VGAscr* scr, int* size, int* align)
{
	Pcidev *p;
	int oapsize, wasupamem;
	ulong aperture, oaperture, fbuf, fbend, *rp;

	oaperture = scr->aperture;
	oapsize = scr->apsize;
	wasupamem = scr->isupamem;

	aperture = 0;
	if(p = sispci()){
		aperture = p->mem[0].bar & ~0x0F;
		*size = p->mem[0].size;
	}

	if(wasupamem){
		if(oaperture == aperture)
			return oaperture;
		upafree(oaperture, oapsize);
	}
	scr->isupamem = 0;

	aperture = upamalloc(aperture, *size, *align);
	if(aperture == 0){
		if(wasupamem && upamalloc(oaperture, oapsize, 0)){
			aperture = oaperture;
			scr->isupamem = 1;
		}
		else
			scr->isupamem = 0;
	}
	else
		scr->isupamem = 1;

	return aperture;
}

static void
sisenable(VGAscr* scr)
{
	Pcidev *p;
	int align, size, tmp, i;
	ulong aperture;

	/*
	 * Only once, can't be disabled for now.
	 * scr->io holds the physical address of
	 * the MMIO registers.
	 */
	if(scr->io)
		return;
	p = sispci();
	if(p == nil)
		return;

	scr->id = p->did;

	scr->io = upamalloc(p->mem[1].bar & ~0x0F, p->mem[1].size, 0);
	if(scr->io == 0)
		return;

	addvgaseg("sismmio", (ulong)scr->io, p->mem[1].size);

	size = p->mem[0].size;
	align = 0;
	aperture = sislinear(scr, &size, &align);
	if(aperture){
		scr->aperture = aperture;
		scr->apsize = size;
		addvgaseg("sisscreen", aperture, size);
	}

	/* find video memory size */
	switch(scr->id){
	case 0x6300:	/* 630 */
		p = pcimatchtbdf(MKBUS(BusPCI, 0, 0, 0));
		tmp = pcicfgr8(p, 0x63);
		if(tmp & 0x80){
			i = (1 << (((tmp & 0x70) >> 4) + 21));
			tmp = pcicfgr8(p, 0x64);
			if(tmp&0x30 == 0x30)
				i <<= 1;
		}else{
			print("sis300: shared memory is disabled\n");
			tmp = vgaxi(Grx, 0x14) | 0xFF;
			i = ((tmp & 0x3F) +1) * MB;
		}
		break;
	}
	scr->storage = i - 4096;
}

VGAdev vgasisdev = {
	"sis",

	sisenable,
	nil,
	nil,
	sislinear,
};

VGAcur vgasiscur = {
	"sis",
};
